#! /usr/bin/env python3

"""Standalone helper script to re-generate the SM covariance data files."""

import argparse
import logging
import sys
logging.basicConfig(level=logging.INFO)

DEFAULT_ckm_scheme = 'CKMSchemeRmuBtaunuBxlnuDeltaM'

def main(argv):
    parser = argparse.ArgumentParser(description='Recompute and save the SM covariances to the working directory.')
    parser.add_argument('-n', type=int, default=5000,
                        help='Number of evaluations (default 5000)')
    parser.add_argument('-t', type=int, default=1,
                        help='Number of threads (default 1)')
    parser.add_argument('-f', action='store_true',
                        help='Force recomputation (default false)')
    parser.add_argument('-s', type=str, default=DEFAULT_ckm_scheme,
                        help="Name of CKM scheme (default {})".format(
                        DEFAULT_ckm_scheme))
    parser.add_argument('--fix_ckm', action='store_true',
                        help='Fix CKM values to their SM values (default false)')

    args = parser.parse_args()

    from smelli import GlobalLikelihood
    gl = GlobalLikelihood(ckm_scheme=args.s, fix_ckm=args.fix_ckm)

    logging.info("Computing covariances with N={} and {} threads".format(args.n, args.t))

    gl.make_measurement(N=args.n, threads=args.t, force=args.f)
    gl.save_sm_covariances('.')


if __name__ == '__main__':
    main(sys.argv)



import flavio
import numpy as np
from math import sin, cos, sqrt, acos
import inspect
from abc import ABC, abstractmethod


m = flavio.Measurement('CKM ratio measurements')
m.set_constraint('RKpi(P+->munu)', '1.3368+-0.0032')
m.set_constraint('DeltaM_d/DeltaM_s', '0.02852+-0.00011')


def get_ckm_schemes():
    return {
        k: v() for k,v in globals().items()
        if inspect.isclass(v)
        and issubclass(v, CKMScheme)
        and len(v.observables) == 4
    }


class CKMScheme(ABC):
    """Base class for schemes to determine the SMEFT CKM elements from
    a set of four input observables.

    This is an abstract class that contains the following abstract methods to be
    overwritten by its subclasses:
    - ckm_fac: static method with arguments `Vus`, `Vcb`, `Vub`, and
      `delta` that returns a list containing the CKM prefactors of the
      four observables in the order given by the `observales` attribute.
    - get_ckm: static method that is the inverse of the `ckm_fac` method
      such that `get_ckm(ckm_fac(Vus, Vcb, Vub, delta))` will return the
      list `[Vus, Vcb, Vub, delta]`.
    - jacobian: static method with arguments `Vus`, `Vcb`, `Vub`, and
      `delta` that returns the Jacobian of the transformation defined by
      the `ckm_fac` method in terms of a numpy array of shape (4,4).
    Furthermore, subclasses have to overwrite the following class attribute:
    - observables: class attribute containing a list of exactly four
      observable strings. Currently, only observables without arguments
      are supported. They must have existing, uncorrelated experimental
      measurements.
    """

    observables = []

    def __init__(self, par_obj=None):
        """Initialize the class.

        Parameters:
        - par_obj: instance of `flavio.classes.ParameterConstraints`.
          Defaults to `flavio.default_parameters`.
        """
        assert len(self.observables) == 4, "Exactly 4 observables should be specified"
        self.par_obj = par_obj or flavio.default_parameters
        # central parameter values
        self.par_central = self.par_obj.get_central_all()
        # central CKM values
        self.ckm_par = ['Vus', 'Vub', 'Vcb', 'delta']
        self.ckm_initial = {k: self.par_central[k] for k in self.ckm_par}

    @staticmethod
    @abstractmethod
    def ckm_fac(Vus, Vcb, Vub, delta):
        pass

    @staticmethod
    @abstractmethod
    def get_ckm(ckm_fac):
        pass


    @staticmethod
    @abstractmethod
    def jacobian(Vus, Vcb, Vub, delta):
        pass

    def sm_covariance(self, N=1000):
        """Compute the covariance of theory predictions for the four input
        observables in the SM, fixing the CKM elements."""
        par_vary = [p for p in self.par_obj.all_parameters if p not in self.ckm_par]
        return flavio.sm_covariance(self.observables, N=N, par_obj=self.par_obj, par_vary=par_vary)

    def exp_measurements(self):
        """Return a list of four probability distributions, corresponding
        to the  experimental measurements of the four input observables."""
        return {obs: flavio.combine_measurements(obs) for obs in self.observables}

    def exp_covariance(self, measurements):
        """Return the covariance of experimental measurements.
        Correlations are currently neglected!
        `measurements` should be the output returned by the `exp_measurements`
        method."""
        return np.diag([measurements[obs].standard_deviation**2 for obs in self.observables])

    def obs_covariance(self, N=1000):
        """Return the covariance (theory and experiment combined) for
        the four input observables in the SM."""
        sm_cov = self.sm_covariance(N=N)
        measurements = self.exp_measurements()
        exp_cov = self.exp_covariance(measurements)
        return sm_cov + exp_cov

    def np_predictions_nockm(self, w=None, **ckm):
        """Return the predictions for the four input observables in the presence
        of NP (parametrized by a `wilson.Wilson` instance `w`) with the
        CKM prefactor removed."""
        fac = self.ckm_fac(**ckm)
        par_dict = self.par_central.copy()
        par_dict.update(ckm)
        if w is None:
            w = flavio.WilsonCoefficients()
        return [flavio.classes.Observable[obs].prediction_par(par_dict, w) / fac[i]
                for i, obs in enumerate(self.observables)]

    def ckm_covariance(self, N=1000):
        """Return the covariance for the four CKM parameters
        `Vus`, `Vcb`, `Vub`, `delta`"""
        ckm = self.ckm_initial
        J = self.jacobian(**ckm)
        cov = self.obs_covariance(N=N)
        pred_sm = self.np_predictions_nockm(w=None, **ckm)
        cov = cov / np.outer(pred_sm, pred_sm)
        iJ = np.linalg.inv(J)
        return iJ @ cov @ iJ.T

    def ckm_fac_np(self, w, **ckm):
        """Return the central values for the four CKM prefactors
        in the presence of new physics (parametrized by a `wilson.Wilson`
        instance `w`) """
        exp = self.exp_measurements()
        exp_cen = np.array([exp[obs].central_value for obs in self.observables])
        np_cen = np.array(self.np_predictions_nockm(w, **ckm))
        return exp_cen / np_cen

    def _ckm_np(self, w=None, **ckm):
        """Return the central values for the four CKM parameters
        `Vus`, `Vcb`, `Vub`, `delta`,
        in the presence of new physics (parametrized by a `wilson.Wilson`
        instance `w`).

        This is a private method where the CKM paramerters in the `Wilson`
        instance and in `**ckm` must be provided manually. For an iterative
        version where this is done automatically, see `ckm_np`."""
        ckm_fac = self.ckm_fac_np(w, **ckm)
        return self.get_ckm(ckm_fac)

    def ckm_np(self, w=None, iterate=10):
        """Return the central values for the four CKM parameters
        `Vus`, `Vcb`, `Vub`, `delta`,
        in the presence of new physics (parametrized by a `wilson.Wilson`
        instance `w`) """
        Vus, Vcb, Vub, delta = [self.ckm_initial[p] for p in ['Vus', 'Vcb', 'Vub', 'delta']]
        for _ in range(iterate):
            if w is None:
                Vus, Vcb, Vub, delta = self._ckm_np(w=None, Vus=Vus, Vcb=Vcb, Vub=Vub, delta=delta)
            else:
                w.set_option('parameters', {'Vus': Vus, 'Vcb': Vcb, 'Vub': Vub, 'gamma': delta})
                Vus, Vcb, Vub, delta = self._ckm_np(w, Vus=Vus, Vcb=Vcb, Vub=Vub, delta=delta)
        return Vus, Vcb, Vub, delta


class CKMSchemeRmuBtaunuBxlnuDeltaM(CKMScheme):
    """CKM scheme where the four input observables are given by:
    - 'RKpi(P+->munu)' (mostly fixing `Vus`)
    - 'BR(B->Xcenu)' (fixing `Vcb`)
    - 'BR(B+->taunu)' (fixing `Vub`)
    - 'DeltaM_d/DeltaM_s' (mostly fixing `delta`)
    """

    observables=[
        'RKpi(P+->munu)',
        'BR(B->Xcenu)',
        'BR(B+->taunu)',
        'DeltaM_d/DeltaM_s'
    ]

    @staticmethod
    def ckm_fac(Vus, Vcb, Vub, delta):
        """Return the for CKM prefactors as function of the four CKM elements"""
        return [
            (Vus**2 / (1 - Vub**2 - Vus**2)),
            Vcb**2,
            Vub**2,
            -((Vcb**2*Vus**2 + Vub**2*(-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2) -
            2*Vcb*Vub*Vus*((-1 + Vcb**2 + Vub**2)**2*(-1 + Vub**2 + Vus**2)**2)**0.25*cos(delta))/
            (Vcb**2*(-1 + Vub**2) + (Vcb**2 + (-1 + Vcb**2)*Vub**2 + Vub**4)*Vus**2 -
            2*Vcb*Vub*Vus*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*cos(delta)))
        ]

    @staticmethod
    def get_ckm(ckm_fac):
        """Inverse of the `ckm_fac` method: returns the four CKM parameters
        given the four CKM prefactors."""
        f1 = ckm_fac[0]  # Vus**2 / Vud**2
        f2 = ckm_fac[1]  # Vcb**2
        f3 = ckm_fac[2]  # Vub**2
        f4 = ckm_fac[3]  # abs(Vtb * Vts)**2 / abs(Vtb * Vtd)**2
        return [
            sqrt(f1 - f1 * f3) / sqrt(1 + f1),  # Vus
            sqrt(f2),  # Vcb
            sqrt(f3),  # Vub
            # delta
            acos(-((-1 + f3)*(f3 - f3*(f2 + f3) - f2*f4 + f1*(f2 + f3*(-1 + f2 + f3)*f4)))/
            (2.*sqrt(1 + f1)*sqrt(f2)*sqrt(f3)*sqrt(f1 - f1*f3)*
            ((((-1 + f3)**2*(-1 + f2 + f3)**2)/(1 + f1)**2)**0.25 + sqrt(((-1 + f3)*(-1 + f2 + f3))/(1 + f1))*f4)))
        ]

    @staticmethod
    def jacobian(Vus, Vcb, Vub, delta):
        """Return the Jacobian of the transformation from the four CKM
        parameters to the four CKM prefactors."""
        J = np.zeros((4, 4))
        J[0, 0] = (-2*(-1 + Vub**2)*Vus)/(-1 + Vub**2 + Vus**2)**2
        J[0, 2] = (2*Vub*Vus**2)/(-1 + Vub**2 + Vus**2)**2
        J[1, 1] = 2 * Vcb
        J[2, 2] = 2 * Vub
        J[3, 0] = ((2*(-1 + Vub**2)**2*(Vcb**2 + Vub**2)*((Vcb**2 + (-1 + Vcb**2)*Vub**2 + Vub**4)*Vus*
            sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2)) - Vcb*Vub*(-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + 2*Vus**2)*cos(delta))
            )/(sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*
            (Vcb**2*(-1 + Vub**2) + (Vcb**2 + (-1 + Vcb**2)*Vub**2 + Vub**4)*Vus**2 -
            2*Vcb*Vub*Vus*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*cos(delta))**2))
        J[3, 1] = ((2*Vub*(-1 + Vub**2)**2*(Vcb*Vub*(-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + 2*Vus**2) -
            (Vcb**2 + (-1 + Vcb**2)*Vub**2 + Vub**4)*Vus*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*cos(delta)))/
            ((-1 + Vcb**2 + Vub**2)*(Vcb**2*(-1 + Vub**2) + (Vcb**2 + (-1 + Vcb**2)*Vub**2 + Vub**4)*Vus**2 -
            2*Vcb*Vub*Vus*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*cos(delta))**2))
        J[3, 2] = ((2*(-1 + Vub**2)*(-(Vub*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*
            (Vcb**2*(-1 + Vub**2)*(-1 + Vcb**2 + 2*Vub**2) +
            (-Vub**4 + Vub**6 + Vcb**4*(3 + Vub**2) + 2*Vcb**2*(-1 + 2*Vub**2 + Vub**4))*Vus**2)) +
            Vcb*Vus*(Vcb**2*(1 - 6*Vub**4 + 5*Vub**6 + (-1 - 2*Vub**2 + 7*Vub**4)*Vus**2) +
            Vub**2*(-1 + Vub**2)*(1 + 3*Vub**4 - Vus**2 + 4*Vub**2*(-1 + Vus**2)) +
            Vcb**4*(-1 + 2*Vub**4 + Vus**2 + Vub**2*(-1 + 3*Vus**2)))*cos(delta)))/
            (sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*
            (Vcb**2*(-1 + Vub**2) + (Vcb**2 + (-1 + Vcb**2)*Vub**2 + Vub**4)*Vus**2 -
            2*Vcb*Vub*Vus*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*cos(delta))**2))
        J[3, 3] = ((2*Vcb*Vub*(-1 + Vub**2)**2*(Vcb**2 + Vub**2)*Vus*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*sin(delta))/
            (Vcb**2*(-1 + Vub**2) + (Vcb**2 + (-1 + Vcb**2)*Vub**2 + Vub**4)*Vus**2 -
            2*Vcb*Vub*Vus*sqrt((-1 + Vcb**2 + Vub**2)*(-1 + Vub**2 + Vus**2))*cos(delta))**2)
        return J



import pkgutil
import os
import sys
from collections import defaultdict


def tree():
    """Tree data structure.

    See https://gist.github.com/hrldcpr/2012250
    """
    return defaultdict(tree)


def get_datapath(package, resource):
    """Rewrite of pkgutil.get_data() that just returns the file path.

    Taken from https://stackoverflow.com/a/13773912"""
    loader = pkgutil.get_loader(package)
    if loader is None or not hasattr(loader, 'get_data'):
        return None
    mod = sys.modules.get(package) or loader.load_module(package)
    if mod is None or not hasattr(mod, '__file__'):
        return None
    # Modify the resource name to be compatible with the loader.get_data
    # signature - an os.path format "filename" starting with the dirname of
    # the package's __file__
    parts = resource.split('/')
    parts.insert(0, os.path.dirname(mod.__file__))
    resource_name = os.path.join(*parts)
    return resource_name



import unittest
from .classes import *


class TestGlobalLikelihood(unittest.TestCase):

    def test_init(self):
        gl = GlobalLikelihood()
        # check that the SM delta ll is 0
        ll_sm = gl.parameter_point({}, 100)
        for k, v in ll_sm.log_likelihood_dict().items():
            self.assertEqual(v, 0, msg="Failed for {}".format(k))
        # check just that this does not give an error
        ll_np = gl.parameter_point({'lq1_2223': 1e-8}, 1000)

    def test_dispatch(self):
        wc_dict  = {'phiD': 1e-10, 'lq1_3323': 1e-7}
        gl = GlobalLikelihood()
        pp = []
        pp.append(gl.parameter_point(wc_dict, 1e3))  # dict, int
        pp.append(gl.parameter_point(wc_dict, scale=1e3))  # dict, kwarg
        pp.append(gl.parameter_point(wc_dict, 1.e3))  # dict, float
        w = Wilson(wc_dict, 1e3, 'SMEFT', 'Warsaw')
        pp.append(gl.parameter_point(w))  # dict, float
        filename = get_datapath('smelli', 'data/test/wcxf.yaml')
        pp.append(gl.parameter_point(filename))  #str
        for i, p in enumerate(pp):
            self.assertDictEqual(wc_dict, p.w.wc.dict,
                                 msg="Failed for {}".format(i))
            self.assertEqual(1e3, p.w.wc.scale,
                             msg="Failed for {}".format(i))

    def test_basis(self):
        gl = GlobalLikelihood(basis='Warsaw up')
        self.assertEqual(gl.basis, 'Warsaw up')

    def test_wet(self):
        gl = GlobalLikelihood(eft='WET')
        # check that the SM delta ll is 0
        ll_sm = gl.parameter_point({}, 100)
        for k, v in ll_sm.log_likelihood_dict().items():
            self.assertEqual(v, 0, msg="Failed for {}".format(k))
        # check just that this does not give an error
        ll_np = gl.parameter_point({'C9_bsmumu': -1}, 4.8)

    def test_incl_excl(self):
        gl = GlobalLikelihood(eft='WET', basis='flavio', include_likelihoods=['fast_likelihood_quarks.yaml'])
        ll = gl.parameter_point({}, 100).log_likelihood_dict()
        self.assertSetEqual(set(ll.keys()), {'fast_likelihood_quarks.yaml', 'global'})
        gl = GlobalLikelihood(eft='WET', basis='flavio', include_likelihoods=['likelihood_lfv.yaml'])
        ll = gl.parameter_point({}, 100).log_likelihood_dict()
        self.assertSetEqual(set(ll.keys()), {'likelihood_lfv.yaml', 'global'})
        gl = GlobalLikelihood(eft='WET', basis='flavio', exclude_likelihoods=['likelihood_lfv.yaml'])
        ll = gl.parameter_point({}, 100).log_likelihood_dict()
        self.assertNotIn('likelihood_lfv.yaml', set(ll.keys()))
        self.assertRaises(ValueError, GlobalLikelihood, include_likelihoods=["nonexistent_likelihood.yaml"])
        self.assertRaises(ValueError, GlobalLikelihood, exclude_likelihoods=["nonexistent_likelihood.yaml"])


class TestGlobalLikelihoodPoint(unittest.TestCase):

    def test_obstable(self):
        gl = GlobalLikelihood()
        res = gl.parameter_point({'lq1_2223': 1e-8}, 91.1876)
        self.assertIsInstance(res, GlobalLikelihoodPoint)
        df = res.obstable(min_pull_exp=1)
        self.assertTrue(df['pull exp.'].min() >= 1)

    def test_pvalue(self):
        gl = GlobalLikelihood()
        res = gl.parameter_point({}, 91.1876)
        pval = res.pvalue_dict()
        self.assertTrue(0 < pval['global'] < 1)



from . import ckm
import unittest
import flavio
from wilson import Wilson
from math import sqrt
import smelli
from cmath import exp


par = flavio.default_parameters.get_central_all()


class TestCKM(unittest.TestCase):
    def test_sm(self):
        scheme = ckm.CKMSchemeRmuBtaunuBxlnuDeltaM()
        scheme.ckm_covariance()
        Vus, Vcb, Vub, delta = scheme.ckm_np(None)
        self.assertAlmostEqual(Vus, 0.225, delta=0.001)
        self.assertAlmostEqual(Vcb, 0.041, delta=0.0015)
        self.assertAlmostEqual(Vub, 0.004, delta=0.0005)
        self.assertAlmostEqual(delta, 1.15, delta=0.2)

    def test_ckm_np(self):
        scheme = ckm.CKMSchemeRmuBtaunuBxlnuDeltaM()
        w = Wilson({'lq3_1123': 0.04e-5,
                    'lq3_3313': 0.008e-5,
                    'lq3_2212': 0.5e-7,
                    'qq1_2323': 1e-12,
                    }, 91.1876, 'SMEFT', 'Warsaw')
        Vus, Vcb, Vub, delta = scheme.ckm_np(w, iterate=10)
        w.set_option('parameters', {'Vus': Vus, 'Vcb': Vcb, 'Vub': Vub, 'gamma': delta})
        _Vus, _Vcb, _Vub, _delta = scheme._ckm_np(w, Vus=Vus, Vcb=Vcb, Vub=Vub, delta=delta)
        self.assertAlmostEqual(_Vus, Vus, places=2)
        self.assertAlmostEqual(_Vcb, Vcb, places=2)
        self.assertAlmostEqual(_Vub, Vub, places=2)
        self.assertAlmostEqual(_delta, delta, places=2)


class TestSmelliCKM(unittest.TestCase):
    def test_init(self):
        gl = smelli.GlobalLikelihood()
        # with fix_ckm
        gl_fixckm = smelli.GlobalLikelihood(fix_ckm=True)
        self.assertEqual(gl.par_dict_default['Vcb'], par['Vcb'])
        VcbSM = gl.par_dict_sm['Vcb']
        VubSM = gl.par_dict_sm['Vub']
        VusSM = gl.par_dict_sm['Vus']
        deltaSM = gl.par_dict_sm['delta']
        self.assertAlmostEqual(par['Vcb'], VcbSM, delta=0.0002)
        self.assertAlmostEqual(par['Vub'], VubSM, delta=0.0005)
        self.assertAlmostEqual(par['Vus'], VusSM, delta=0.0006)
        pre = -4 * par['GF'] / sqrt(2)
        # Vcb
        w = Wilson({'lq3_1123': 0.5 * pre * VcbSM * (-0.5)}, 91.1876, 'SMEFT', 'Warsaw')
        pp = gl.parameter_point(w)
        self.assertAlmostEqual(pp.par_dict_np['Vcb'] / VcbSM,  1.5, delta=0.03)
        # with fix_ckm
        pp = gl_fixckm.parameter_point(w)
        self.assertEqual(pp.par_dict_np['Vcb'] / par['Vcb'],  1)
        # Vub
        w = Wilson({'lq3_3313': 0.5 * pre * VubSM * (-0.5) * exp(-1j * deltaSM)}, 91.1876, 'SMEFT', 'Warsaw')
        pp = gl.parameter_point(w)
        self.assertAlmostEqual(pp.par_dict_np['Vub'] / VubSM, 1.5, delta=0.03)
        # Vus
        w = Wilson({'lq3_2212': 0.5 * pre * VusSM * (-0.1)}, 91.1876, 'SMEFT', 'Warsaw')
        pp = gl.parameter_point(w)
        self.assertAlmostEqual(pp.par_dict_np['Vus'] / VusSM, 1.1, delta=0.03)

    def test_fast_likelihoods(self):
        scheme = ckm.CKMSchemeRmuBtaunuBxlnuDeltaM()
        ckm_central = scheme.ckm_np()
        gl = smelli.GlobalLikelihood()
        for fl in gl.fast_likelihoods.values():
            par = fl.par_obj
            self.assertAlmostEqual(par.get_central('Vus'), ckm_central[0], delta=0.00001)
            self.assertAlmostEqual(par.get_central('Vcb'), ckm_central[1], delta=0.00001)
            self.assertAlmostEqual(par.get_central('Vub'), ckm_central[2], delta=0.00001)
            self.assertAlmostEqual(par.get_central('delta'), ckm_central[3], delta=0.0001)



from ._version import __version__, __flavio__version__
from . import classes
from . import util
from .classes import GlobalLikelihood
import flavio
import warnings


def _check_flavio_version():
    version_installed = flavio.__version__.split('.')
    version_required = __flavio__version__.split('.')
    if version_installed[:2] == version_required[:2]:
        return True
    else:
        return False


if not _check_flavio_version():
    warnings.warn("Your installed flavio version ({})"
                  .format(flavio.__version__)
                  + " is not compatible with this version of smelli."
                  " Please check for available updates.")



import smelli
from smelli.ckm import get_ckm_schemes
import numpy as np
import yaml

FILE = 'par_ckm'

def save_ckm(scheme,file_path,N=10000):
    val = scheme.ckm_np()
    cov = scheme.ckm_covariance(N=10000)
    err = np.sqrt(np.diag(cov))
    corr = cov / np.outer(err, err)

    dat = [
      {
        "parameters": [
          "Vus",
          "Vcb",
          "Vub",
          "delta"
        ],
        "values": {
          "distribution": "multivariate_normal",
          "central_value": list(val),
          "standard_deviation": err.tolist(),
          "correlation": corr.tolist()
        }
      }
    ]

    with open(file_path, 'w') as f:
        yaml.dump(dat, f)

for name, ckm_scheme in get_ckm_schemes().items():
    file_path = '{}_{}.yaml'.format(FILE,name)
    save_ckm(ckm_scheme,file_path,N=10000)
